from skilletlib import Skillet


class BaseCommand:
    sli_command = 'unimplemented'

    def __init__(self, sli):
        self.sli = sli
        self.args = self.sli.args
        # most commands use a skillet by default, subclasses should override as necessary
        self.no_skillet = False

    def execute(self):
        """SLI starts here to call the run function"""

        # Filter out all var=value arguments for context loading
        self.raw_args = self.args
        self.args = [x for x in self.args if "=" not in x]

        self.run()
        self._get_output()

    def execute_debug(self):
        self.raw_args = self.args
        self.args = [x for x in self.args if "=" not in x]
        if hasattr(self, 'debug'):
            self.debug()
        else:
            print(f"Module {self.sli_command} doesn't have a debug function")

    def run(self):
        """Should be overridden by child classes"""
        raise Exception(f"The SLI module {self.sli_command} has not properly overridden the run function")

    def _print_usage(self):
        if hasattr(self, "help_text"):
            print(self.help_text)
        elif hasattr(self, "short_desc"):
            print(self.short_desc)
        else:
            print(f"{self.sli_command} does not provide any help")

    def _get_output(self):

        if not self.no_skillet:
            skillet: Skillet = self.sli.skillet

            if not skillet:
                return

            results = skillet.get_results()

            if "output_template" in results:
                print(results["output_template"])

            elif self.sli.verbose:
                print(results)

            elif not getattr(self, "suppress_output", False) is True:
                print(f"Result Success: {skillet.success}")



import json

from skilletlib.exceptions import PanoplyException

from sli.decorators import load_variables
from sli.decorators import require_ngfw_connection_params
from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type
from .base import BaseCommand


class ConfigureCommand(BaseCommand):
    sli_command = "configure"
    short_desc = "Execute a configuration skillet of type panos"
    help_text = """
    Executes a PAN-OS configuration skillet.

    Example: Load and commit and configuration skillet
        sli configure --name test_skillet -sd ~/pan/Skillets -uc

    Example: Verify the actions of a configuration skillet ONLY
        sli configure --debug --name test_skillet -sd ~/pan/Skillets -uc

"""

    @require_single_skillet
    @require_skillet_type("panos", "panorama")
    @require_ngfw_connection_params
    @load_variables
    def run(self):

        print(f"Executing {self.sli.skillet.name}...")
        output = self.sli.skillet.execute(self.sli.context)

        needs_commit = output.get("changed", False)

        if needs_commit and self.sli.commit:
            print("Committing configuration...")
            self.sli.skillet.panoply.commit()
            print("Finished")

        else:
            if needs_commit:
                print("Configuration loaded into candidate config")

            else:
                context_keys = output.get("outputs", {}).keys()

                if len(context_keys):
                    print("Saved the following items into the context:")
                    print("----")

                    for k in context_keys:
                        print(k)

                    print("----")

    @require_single_skillet
    @require_skillet_type("panos", "panorama")
    @require_ngfw_connection_params
    @load_variables
    def debug(self):
        """
        Debug this configuration skillet, this will not perform any destructive actions against the device, but
        will perform op commands and the like to gather information
        """
        print(f"Running {self.sli.skillet.name} debug...")
        skillet_context = self.sli.skillet.initialize_context(self.sli.context)
        changes = dict()

        for snippet in self.sli.skillet.get_snippets():
            snippet.update_context(skillet_context)

            loop_vars = snippet.get_loop_parameter()
            index = 0

            for item in loop_vars:
                change = dict()

                skillet_context["loop"] = item
                skillet_context["loop_index"] = index

                snippet.render_metadata(skillet_context)

                if snippet.name in changes:
                    changes[f"{snippet.name}_{index}"] = change
                else:
                    changes[snippet.name] = change

                change["metadata"] = json.dumps(snippet.metadata, indent=4)
                change["when"] = True

                if not snippet.should_execute(skillet_context):
                    change["message"] = "This snippet would be skipped due to when conditional"
                    change["when"] = False
                    continue

                if "cmd" in snippet.metadata and snippet.metadata["cmd"] in (
                    "op",
                    "set",
                    "edit",
                    "override",
                    "move",
                    "rename",
                    "clone",
                    "delete",
                ):

                    change["message"] = "This snippet would be executed"

                else:
                    try:
                        (output, status) = snippet.execute(skillet_context)
                        # capture all outputs
                        snippet_outputs = snippet.get_default_output(output, status)
                        captured_outputs = snippet.capture_outputs(output, status)

                        skillet_context.update(snippet_outputs)
                        skillet_context.update(captured_outputs)

                        change["message"] = "This snippet was executed to gather results"
                        change["captured_output"] = json.dumps(captured_outputs, indent=4)

                    except PanoplyException as pe:
                        change["message"] = str(pe)

                index = index + 1
                snippet.reset_metadata()

        print("****Debug Output ****")
        for k, v in changes.items():
            print(f"Snippet: {k}")
            if isinstance(v, dict):
                for kk, vv in v.items():
                    try:
                        j = json.loads(vv)
                        print("{")
                        for jk, jv in j.items():
                            print(f"   {jk}: {jv}")
                        print("}")
                    except (TypeError, json.decoder.JSONDecodeError):
                        print(f"{kk}: {vv}")

                print("***\n")
            else:
                print(f"{k}\n{v}\n")



from .base import BaseCommand

from ..decorators import require_single_skillet


class ShowSkillet(BaseCommand):
    sli_command = "show_skillet"
    short_desc = "Shows the contents of a Compiled Skillet in YAML format"
    no_skillet = False
    no_context = True
    help_text = """

    Usage:
        sli show_skillet --name k12_config_skillet -sd /tmp/skillets
    """

    @require_single_skillet
    def run(self):
        print(self.sli.skillet.dump_yaml())



from .base import BaseCommand
from sli.async_ssh import AsyncSSHSession
from sli.tools import print_table
import asyncio
from asyncssh.misc import PermissionDenied
import re
import yaml
from getpass import getpass
import os


class MassSSH(BaseCommand):
    sli_command = "mass_ssh"
    short_desc = "Run a command script against multiple firewalls async"
    no_skillet = True
    help_text = """
        Execute a script of CLI commands against multiple firewalls asynchronously. The var
        script.txt must reference a file that contains the CLI commands to run, and devices
        must be either a comma separated list of devices, or a configuration file describing
        connectivity.

        The -o option refers to a directory to create and populate with output logs from all
        devices configured. The contents of the directory will be overwritten if it already exists.

        Providing credentials in the yaml file is optional and may be passed from the CLI,
        however the yaml file provides support for overriding credentials for specific devices.

        Sample structuring example of the yaml file:

        ---

        creds:
            username: global_user
            password: global_password

        devices:
            - device: device_one

            - device: device_two
              username: device_two_user
              password: device_two_password

        ---

        Sample structuring of a script.txt configuring a new zone

        ---

        configure
        set zone new_zone
        commit

        ---

        the input script is not limited to supporting just configuration commands, show commands can be used for
        data gathering on a list of devices simultaneously

        Usage:
            sli mass_ssh -u username -p password -o output_dir script.txt [comma-seperated-devices | config.yaml]
    """

    def get_credentials(self):
        """
        Helper function to first check options and context for credentials, then prompt user if required
        """
        username = self.sli.options.get("username", self.sli.context.get("TARGET_USERNAME", ""))
        password = self.sli.options.get("password", self.sli.context.get("TARGET_PASSWORD", ""))
        while not len(username):
            username = input("Default username: ")
        while not len(password):
            password = getpass("Default password: ")
        return username, password

    def load_config_from_yaml(self, out_directory):
        """
        Load list of devices and optional credentials from specified config.yaml file
        """
        # Assemble devices from YAML configuration file
        with open(self.args[1], "r") as f:
            devices = yaml.safe_load(f)
        # Check for presence of devices without specific credentials
        if len([x for x in devices["devices"] if "username" not in x and "password" not in x]):

            # Get default credentials if global credentials not specified
            username = None
            password = None
            if "creds" in devices:
                username = devices["creds"].get("username")
                password = devices["creds"].get("password")
            if not username or not password:
                username, password = self.get_credentials()

            # Apply default parameters to all devices
            devices = devices["devices"]
            for device in devices:
                if "username" not in device:
                    device["username"] = username
                if "password" not in device:
                    device["password"] = password
                device["out_file"] = f"{os.path.join(out_directory, device['device'])}.txt" if out_directory else None
                device["coroutine"] = None
                device["status"] = False
                device["error"] = "",
                device["output"] = None

        return devices

    def load_config_from_list(self, out_directory):
        """
        Create configuration object based on users CLI input
        """
        username, password = self.get_credentials()
        return [{
                    "device": x,
                    "username": username,
                    "password": password,
                    "out_file": f"{os.path.join(out_directory,x)}.txt" if out_directory else None,
                    "coroutine": None,
                    "status": False,
                    "output": None,
                    "error": "",
                } for x in self.args[1].split(",")]

    @staticmethod
    async def ssh_coroutine(device, username, password, out_file, script, dev_obj):
        """
        Per device coroutine
        """
        client = AsyncSSHSession(device, username, password)
        try:
            await client.connect()
            dev_obj["output"] = await client.run_command_script(script, out_file=out_file)
            if client.has_error:
                dev_obj["error"] = client.error
            else:
                dev_obj["status"] = True
        except OSError:
            dev_obj["error"] = "Unable to connect to device"
        except PermissionDenied:
            dev_obj["error"] = "Device rejected login"
        except Exception as e:
            dev_obj["error"] = e

    async def mass_ssh(self, devices):
        """
        Run gather on individual coroutines
        """
        # Gather and start coroutines
        tasks = [x["coroutine"] for x in devices]
        print(f"Starting SSH to {len(devices)} devices")
        await asyncio.gather(*tasks, return_exceptions=True)

    def run(self):

        if not len(self.args) == 2:
            print(self.help_text)
            return

        with open(self.args[0], "r") as f:
            script = f.read()

        # Create output directory if doesn't exist
        out_directory = self.sli.options.get("out_file", None)
        if out_directory:
            if not os.path.exists(out_directory):
                os.mkdir(out_directory)

        # Load devices configuration from YAML or CLI input
        devices = None
        if re.match(r".*\.y.*ml$", self.args[1]):
            devices = self.load_config_from_yaml(out_directory)
        else:
            devices = self.load_config_from_list(out_directory)

        # Populate devices objects with coroutines
        for dev in devices:
            dev["coroutine"] = self.ssh_coroutine(
                dev["device"],
                dev["username"],
                dev["password"],
                dev["out_file"],
                script,
                dev
            )

        # Execute SSH sessions
        asyncio.get_event_loop().run_until_complete(self.mass_ssh(devices))

        # If not saving to file, print device output
        if not out_directory:
            for dev in devices:
                if dev["status"]:
                    print(f"{dev['device']}\n{'-'*len(dev['device'])}")
                    print(dev["output"])

        # Print results from all devices
        results = [
                {
                    "device": x["device"],
                    "status": "SUCCESS" if x["status"] else str(x["error"])
                } for x in devices
            ]
        print_table(results, {"Device": "device", "Status": "status"})



from .base import BaseCommand
from sli.decorators import require_single_skillet, require_skillet_type

from lxml import etree
import re
from io import BytesIO


class CreateTemplate(BaseCommand):
    sli_command = 'create_template'
    short_desc = 'Create an XML template from a panos or panorama skillet'
    no_context = True
    help_text = """

    Usage:
        sli create_template -n [skillet] [baseline-file] [out-file]

"""

    @staticmethod
    def insert_template_str(line, config):
        """
        Break up an xml config (str) at a specified line to inject template_string.
        """
        i = 1
        r = ""
        for cl in [x for x in config.split("\n") if len(x)]:
            if line["line"] == i:

                leading_spaces = len(cl) - len(cl.lstrip())
                closing_tag = None

                # If tag without children, expand to tag with children and insert template
                if re.search(r'<[a-zA-Z0-9-]+/>', cl):
                    ll = cl.replace("/", "")
                    r += ll + "\n"
                    closing_tag = ll.replace("<", "</") + "\n"

                # If not a tag at all, there's already a template here, just insert
                elif not cl.strip().startswith("<"):
                    pass

                # If just an opening tag is present, insert after
                else:
                    r += cl + "\n"

                for tl in line["template"].split("\n"):
                    r += " " * leading_spaces + tl + "\n"
                if closing_tag is not None:
                    r += closing_tag

            else:
                r += cl + "\n"
            i += 1
        return r

    @staticmethod
    def find_closest_entry(xml, xpath):
        """
        find the shortest xpath that returns a valid element, and determine what
        nodes are missing from the full expath. Returns a tuple containing the
        shortened xpath that matches a node, and a list of the missing nodes
        """
        xs = [x for x in xpath.split("/") if len(x)]
        for i in range(len(xs)):
            xpath_short = "/" + "/".join(xs[:-1 * (i + 1)])
            if len(xpath_short) < 1:
                raise Exception(f"Could not find valid entry point for {xpath}")
            found = xml.xpath(xpath_short)
            if len(found) == 0:
                continue
            else:
                missing = [x for x in xpath.replace(xpath_short, "").split("/") if len(x)]
                break
        return xpath_short, missing

    @staticmethod
    def strip_var_spaces(in_str) -> str:
        """Remove spaces around variables inside of a block of text"""
        return in_str.replace("{{ ", "{{").replace(" }}", "}}")

    @require_single_skillet
    @require_skillet_type("panos", "panorama")
    def run(self):
        if not len(self.args) == 2:
            print(self.help_text)
            return

        out_file = self.args[1]
        baseline_file = self.args[0]
        with open(baseline_file, "r") as f:
            config = f.read()
            f.seek(0)
            parser = etree.XMLParser(remove_blank_text=True)
            baseline_xml = etree.parse(f, parser)

        snippets = self.sli.skillet.get_snippets()

        # Verify required parameters present in snippets
        for snippet in snippets:
            if getattr(snippet, "template_str", None) is None:
                print(f"Snippet {snippet.name} has no template_str")
                return
            if "xpath" not in snippet.metadata:
                print(f"Snippet {snippet.name} has no xpath")
                return

        # Expand any missing XML nodes
        for snippet in snippets:
            snippet_xpath = self.strip_var_spaces(snippet.metadata["xpath"])
            if len(baseline_xml.xpath(snippet_xpath)):
                continue
            short_xpath, missing = self.find_closest_entry(baseline_xml, snippet_xpath)
            ele = baseline_xml.xpath(short_xpath)[0]
            for missing_ele in missing:
                attr = None
                new_ele_tag = missing_ele
                if "[" in missing_ele:
                    es = missing_ele.split("[")
                    new_ele_tag = es[0]
                    attr = es[1].strip()
                    for c in "[]@'\" ":
                        attr = attr.replace(c, "")
                    attr = attr.split("=")
                new_ele = etree.Element(new_ele_tag)
                if attr is not None:
                    new_ele.set(*attr)
                ele.append(new_ele)
                ele = new_ele

        # Rewrite config var and reload xml document to ensure accurate line numbers
        temp_file = BytesIO()
        baseline_xml.write(temp_file, pretty_print=True)
        temp_file.seek(0)
        config = temp_file.read().decode()
        baseline_xml = etree.fromstring(config)

        # Find the various insert points on all snippets
        lines = []
        for snippet in snippets:

            xpath = snippet.metadata["xpath"]
            found = baseline_xml.xpath(self.strip_var_spaces(xpath))

            if len(found) > 1:
                raise Exception(f"xpath {xpath} returned more than 1 result in baseline")
            elif not len(found):
                raise Exception(f" Unable to find entry point for {xpath}")

            # Insert point found
            else:
                lines.append({
                    "template": self.strip_var_spaces(snippet.template_str),
                    "line": found[0].sourceline
                })

        # Sort the keys so we're starting from the point furthest down the file
        lines = sorted(lines, key=lambda i: i["line"], reverse=True)

        # Insert snippets one at a time until complete
        for line in lines:
            config = self.insert_template_str(line, config)

        with open(out_file, "w") as f:
            f.write(config)



import base64
import hashlib
from Crypto import Random
from Crypto.Cipher import AES
import json

"""
Class to provide encryption and decryption capabilities to various strings throughout project
credit to: https://stackoverflow.com/questions/12524994/encrypt-decrypt-using-pycrypto-aes-256
Modified to accept strings and dictionaries.
"""


class Encryptor(object):

    def __init__(self, key):
        self.bs = AES.block_size
        self.key = hashlib.sha256(key.encode()).digest()

    def encrypt(self, raw):
        """
        Encrypt string raw

        :param raw: string to encrypt
        :type raw: str

        :return: base64 encoded encrypted string
        :rtype: str
        """
        raw = self._pad(raw)
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(raw.encode())).decode('utf-8')

    def decrypt(self, enc):
        """
        Encrypt encrypted string

        :param enc: string of encrypted text to decrypt
        :type enc: str

        :return: decrypted contents of enc string
        :rtype: str
        """
        enc = base64.b64decode(enc.encode())
        iv = enc[:AES.block_size]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(enc[AES.block_size:])).decode('utf-8')

    def encrypt_dict(self, obj):
        """
        Encrypt a dictionary, returning encrypted string encoded in Base64

        :param obj: dictionary object to tbe encrypted
        :type obj: dict

        :return: encrypted string representing obj
        :rtype: str
        """
        return self.encrypt(json.dumps(obj))

    def decrypt_dict(self, enc):
        """
        Decrypt and return a dictionary stored in encrypted string enc

        :param enc: string of encrypted text to decrypt
        :type enc: str

        :return: dict of contents stored in enc
        :rtype: dict
        """
        return json.loads(self.decrypt(enc))

    def _pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s)-1:])]



from sli.tools import expandedHomePath
import os
import json
import yaml
import getpass
from sli.encrypt import Encryptor

"""
SLI's context manager, load, manipulate, and store context objects
"""


class ContextManager():

    def __init__(self, options):
        self._setup_directory()
        self.options = options
        self.use_context = self.options.get('use_context')
        self.encrypt_context = self.options.get('encrypt_context')
        self.context_dir = expandedHomePath('.sli/context')
        self.context_file = ''  # Populated when loading context
        self.context_password = options.get("context_password", "")
        if len(self.context_password) > 0:
            self.encrypt_context = True

    @staticmethod
    def _setup_directory():
        """Create initial directories required for context management in SLI"""
        directories = [expandedHomePath(d) for d in ['.sli', '.sli/skillets', '.sli/context']]
        for directory in directories:
            if not os.path.exists(directory):
                os.mkdir(directory)

    @staticmethod
    def remove_context(context_name):
        """Remove a specified context file"""
        context_file = expandedHomePath(f'.sli/context/{context_name}.json')
        if not os.path.exists(context_file):
            print(f'Specified context {context_name} does not exist')
            return False
        os.remove(context_file)
        return True

    def clean_context(self, context_name):
        """Remove all keys from a contexts except NGFW credentials"""
        context = self.load_context(from_file=context_name)
        creds_strings = "TARGET_IP", "TARGET_USERNAME", "TARGET_PASSWORD"
        keys = [x for x in context.keys() if x not in creds_strings]
        for key in keys:
            context.pop(key)
        self.use_context = True
        self.save_context(context)

    @staticmethod
    def get_contexts():
        contexts = []
        context_path = expandedHomePath('.sli/context/')
        context_files = [x for x in os.listdir(context_path) if x.endswith('.json')]
        for cf in context_files:
            cf_path = expandedHomePath(f'.sli/context/{cf}')
            with open(cf_path, 'r') as f:
                contents = json.loads(f.read())
            contexts.append({'name': cf.replace('.json', ''), 'encrypted': contents.get('encrypted')})
        return contexts

    def _get_context_password(self):
        """Returns user supplied password for context"""
        if len(self.context_password) > 0:
            return self.context_password
        self.context_password = getpass.getpass('Context encryption password: ')
        return self.context_password

    def load_environment(self):
        """Load and return specified environment file from options"""
        context = {}
        environment_file = self.options.get("environment", "")
        if not len(environment_file):
            return context
        with open(environment_file, 'r') as f:
            if environment_file.endswith(".yaml"):
                context = yaml.safe_load(f)
            else:
                context = json.loads(f.read())
        return context

    def load_context(self, from_file=''):
        """Load a context from disk"""
        context = {}

        if len(from_file):
            # If called with from_file, use that
            context_name = from_file
        else:
            # Unless a context name is specified, assume default context
            context_name = self.options.get('context_name', 'default')
        self.context_file = expandedHomePath(f'.sli/context/{context_name}.json')

        if not context_name == 'default' and not len(from_file):
            self.use_context = True

        # If not loading a context, return the context with environment
        if not self.use_context and not len(from_file):
            context.update(self.load_environment())
            return context

        # If specified context file not found, return blank context
        if not os.path.exists(self.context_file):
            return context

        try:
            with open(self.context_file, 'r') as f:
                context_file_json = json.loads(f.read())
        except Exception as e:
            print(f"Error loading context file {self.context_file} - {e}")
            return context

        # Check for encryption and decrypt stored value if found
        if context_file_json.get('encrypted'):
            self.encrypt_context = True  # Make sure we encrypt the written file at the end
            content = context_file_json.get('encrypted_context')
            if not content:
                raise ValueError(
                    'Context has encryption specified, but encrypted_context not found'
                    )
            password = self._get_context_password()
            decrypted = False
            while not decrypted:
                try:
                    encryptor = Encryptor(password)
                    decrypted_dict = encryptor.decrypt_dict(content)
                    decrypted = True
                except (json.decoder.JSONDecodeError, UnicodeDecodeError):
                    print('Invalid context decryption key\n')
                    self.context_password = ''
                    password = self._get_context_password()
            decrypted_dict.update(self.load_environment())
            return decrypted_dict

        # Assume unencrypted context content, return context
        context.update(context_file_json.get('context', context))
        context.update(self.load_environment())
        return context

    def save_context(self, context):
        """Save a context to disk"""

        # If not configured to use a context do nothing
        if not self.use_context:
            return

        # If required write encrypted context file
        if self.encrypt_context:

            if len(self.context_password) < 1:
                # We don't currently have a context password set, lets create one
                matches = False
                new_password = ''
                while not matches:
                    new_password = getpass.getpass('New context encryption password: ')
                    confirm_password = getpass.getpass('Confirm password: ')
                    matches = new_password == confirm_password and len(new_password) > 0
                    if not matches:
                        print('Passwords did not match.\n')
                self.context_password = new_password

            encryptor = Encryptor(self.context_password)
            encrypted_context = encryptor.encrypt_dict(context)
            with open(self.context_file, 'w') as f:
                write_dict = {
                    'encrypted': True,
                    'encrypted_context': encrypted_context
                }
                f.write(json.dumps(write_dict, indent=4))

        else:
            # Write unencrypted context
            with open(self.context_file, 'w') as f:
                write_dict = {
                    'encrypted': False,
                    'context': context
                }
                f.write(json.dumps(write_dict, indent=4))



from skilletlib import Skillet
from skilletlib import SkilletLoader
from inspect import isclass
from sli import commands
from types import ModuleType

from sli.contextManager import ContextManager
from sli.commands.base import BaseCommand


class SkilletLineInterface:

    def __init__(self, options, action, args):
        self.action = action
        self.options = options
        self.args = args
        self._unpack_options()
        self.command_map = {}
        self._load_commands()
        self._verify_command()
        self.sl = SkilletLoader()
        self.no_skillet = getattr(self.command_map[self.action], "no_skillet", False) is True
        self.no_context = getattr(self.command_map[self.action], "no_context", False) is True

        # Load skillets only if the command requires them
        if not self.no_skillet:
            self._load_skillets()
            self._verify_loaded_skillets()

        # Load context if requested, load only a specified environment otherwise
        self.cm = ContextManager(self.options)
        if not self.no_context:
            self.context = self.cm.load_context()
        else:
            self.context = self.cm.load_environment()
        self.skillet: Skillet = None  # Active running skillet

    def _unpack_options(self):
        """Unpack options onto self where required"""
        for opt in ("verbose", "commit", "report", "loader_error", "output_format", "no_config", "context_var"):
            setattr(self, opt, self.options.get(opt, False))
        self.report_file = self.options.get("report_file", "")

        # If a report file was specified, assume we want to create it
        if self.report_file:
            self.generate_report = True

    def _load_commands(self, command_map=None):
        """
        Loads a list of commands from the sli.commands package. Does so by iterating over contents of the
        package and locating types of 'type' that have an sli_command specified.
        """

        if command_map is None:
            command_map = self.command_map

        # Go through each module in the commands package
        for item in dir(commands):
            item_obj = getattr(commands, item)
            if isinstance(item_obj, ModuleType):

                # Get only the attributes which are classes
                for item_attr in dir(item_obj):
                    item_attr_obj = getattr(item_obj, item_attr)
                    if isclass(item_attr_obj):

                        # Only load commands that subclass BaseCommand
                        if item_attr_obj == BaseCommand or not issubclass(item_attr_obj, BaseCommand):
                            continue

                        # Load only SLI command modules that are appropriately written
                        command_string = getattr(item_attr_obj, "sli_command", "")
                        if len(command_string) > 0:
                            command_map[command_string] = item_attr_obj

    def _verify_command(self):
        """Called in __init__ to verify a submitted command is valid before running SkilletLoader"""
        if self.action not in self.command_map:
            print('Invalid action, run "sli --help" for list of available actions')
            exit(1)

    def _load_skillets(self):
        """Called in __init__ to front end SkilletLoader"""
        self.skillets = self.sl.load_all_skillets_from_dir(self.options.get("directory", "./"))
        if len(self.sl.skillet_errors) > 0:
            print("Errors on loading skillets:")
            for err in self.sl.skillet_errors:
                for key in err:
                    print(f"   {key} - {err[key]}")
            if self.loader_error:
                raise Exception("SkilletLoader encountered errors")

    def _verify_loaded_skillets(self):
        """Perform any pre-execution validations against loaded skillets"""

        if len(self.skillets) < 1:
            print("No skillets were loaded.")
            exit(1)

    @classmethod
    def get_commands(self):
        """Load commands for purposes other than execution"""

        command_map = {}
        self._load_commands(self, command_map=command_map)
        return command_map

    def execute(self):
        """Run supplied SLI command"""

        action_obj = self.command_map[self.action](self)
        if self.options.get("debug"):
            action_obj.execute_debug()
        else:
            action_obj.execute()

        # Update context with new keys from skillet run
        if not self.no_context:
            skillet_context = getattr(self.skillet, "context", None)
            if skillet_context:
                for key in skillet_context.keys():
                    if key not in ["loop", "loop_index"]:
                        self.context[key] = skillet_context[key]
            self.cm.save_context(self.context)

        # Clean run, normal exit
        exit(0)



from .base import BaseCommand
from sli.decorators import require_ngfw_connection_params, require_panoply_connection

"""
Revert a configuration on an NGFW
"""


class RevertCommand(BaseCommand):
    sli_command = 'revert'
    short_desc = 'Revert a candidate configuration on an NGFW'
    no_skillet = True

    @require_ngfw_connection_params
    @require_panoply_connection
    def run(self, pan):

        pan.load_config("running-config.xml")
        print("Configuration reverted to running-config")



import os
from os.path import expanduser
from panforge import Report
from getpass import getpass
import socket
from jinja2 import Environment
import jmespath
from lxml import etree
from io import BytesIO
from pathlib import Path
from sli.errors import AppSkilletNotFoundException
from skilletlib import Skillet
from skilletlib import SkilletLoader


def get_var(var, args, context, options={}):
    """Get a var by searching args and the context, else ask the user"""
    args = {x.split('=')[0]: x.split('=')[1] for x in args if '=' in x}
    if var['name'] in args:
        # First order of preference is to use a CLI provided parameter
        context[var['name']] = args[var['name']]
    else:
        # If input has not yet been supplied, get it from the user
        context.update(get_variable_input(var, context, defaults=options.get("defaults", False)))


def render_template(template_text, context):
    """Render a template loaded from a string with an appropriate environment"""
    env = Environment(extensions=['jinja2_ansible_filters.AnsibleCoreFiltersExtension'])
    return env.from_string(template_text).render(context)


def render_expression(expression, context):
    """Shortcut to render an expression, adding {{}} brackets to a template"""
    return render_template("{{" + str(expression) + "}}", context)


def is_ip(ip):
    """
    Return True if either an IPv4 or IPV6 address
    """
    for family in socket.AF_INET, socket.AF_INET6:
        try:
            socket.inet_pton(family, ip)
            return True
        except Exception:
            pass
    return False


def print_table(objs, defs):
    """
    Print a formatted table with left justified columns
    """

    def get_cell(text, max):
        if text is None:
            return " " * max
        return f"  {text + ' ' * (max - len(text))}  "

    # Calculate max width of each column
    cols = {k: {"width": len(k)} for k in defs}
    for o in objs:
        for k in defs:
            if "." in defs[k]:
                text = jmespath.search(defs[k], o)
            else:
                text = o[defs[k]]
            if text is None:
                continue
            if len(text) > cols[k]["width"]:
                cols[k]["width"] = len(text)
    # Print headers
    header = "".join([get_cell(x, cols[x]["width"]) for x in cols])
    print(header)
    print("-" * len(header))

    # Print data
    for o in objs:
        row = ""
        for k in defs:
            if "." in defs[k]:
                text = jmespath.search(defs[k], o)
            else:
                text = o[defs[k]]
            row += get_cell(text, cols[k]["width"])
        print(row)


def generate_report(out_file, data, report_dir, header=None):
    """
    Generate a panforge report at 'out_file' from source 'data' using 'report_dir'
    as the root reporting directory for panforge
    """
    if not os.path.exists(report_dir):
        print(f'Could not generate report, source directory {report_dir} not present')
        return
    report = Report(report_dir)
    if header:
        report.load_header(header)
    report.load_data(data)
    report_html = report.render_html()
    with open(out_file, 'w') as f:
        f.write(report_html)
    print(f'Report written to {out_file}')


def input_yes_no(prompt, ret_value, ret_false=None, default=None):
    """
    Simple wrapper for asking user yes no questions until an acceptable answer
    is received, returns ret_value if yes, ret_false if no
    """
    default_str = " (y/n): "
    if isinstance(default, str):
        default_yes = True if default.lower() in ("y", "yes") else False
        default_str = f" (y/n default: {'yes' if default_yes else 'no'}): "
    while True:
        user_input = input(prompt + default_str)
        if not len(user_input) and default is not None:
            return ret_value if default_yes else ret_false
        if user_input.lower() in ('y', 'yes'):
            return ret_value
        if user_input.lower() in ('n', 'no'):
            return ret_false
        print('Please input either y or n')


def check_default_empty(default):
    """Helper function to validate if a default parameter is empty based on type"""
    if isinstance(default, str) or isinstance(default, list):
        return len(default) == 0
    elif isinstance(default, int):
        return False
    elif isinstance(default, dict):
        return len(default.keys()) == 0


def get_variable_input(var, context, defaults=False):
    """
    Get input from user in reference to var, being a skillet variable object
    """
    type_hint = var.get('type_hint')
    name = var.get('name')
    desc = var.get('description', name)

    # First get a default from the context, otherwise use the skillet definition
    default = context.get(name, "")
    default_is_empty = check_default_empty(default)

    if default_is_empty:
        default = var.get("default", "")
        default_is_empty = check_default_empty(default)

    if defaults and not default_is_empty:
        return {name: default}

    default_str = f"({default})" if not default_is_empty else ""
    if not name:
        raise ValueError('Input variable missing name')

    ret_dict = {}

    # Check for a toggle hint and return if not applicable
    if "toggle_hint" in var:
        if isinstance(var["toggle_hint"]["value"], list):
            if not context.get(var["toggle_hint"].get("source")) in var["toggle_hint"]["value"]:
                return ret_dict
        else:
            if not var["toggle_hint"]["value"] == context.get(var["toggle_hint"].get("source")):
                return ret_dict

    if type_hint == 'checkbox':

        if 'source' in var:
            raw_dd_list = context.get(var["source"], [])
            if not isinstance(raw_dd_list, list):
                raw_dd_list = [raw_dd_list]

            cbx_list = [{"key": x, "value": x} for x in raw_dd_list]

        else:
            cbx_list = var.get('cbx_list', [])

        print(f"\n{desc}\n{'-'*len(desc)}\n")

        # Get inputs from user until user confirms they are ok
        confirmed = False
        input_list = []
        while not confirmed:
            input_list.clear()
            for cbx in cbx_list:
                default_result = None
                if isinstance(default, list):
                    default_result = "y" if cbx["value"] in default else "n"
                input_value = input_yes_no(cbx['key'], cbx['value'], default=default_result)
                if input_value:
                    input_list.append(input_value)

            # Display all inputs
            check = []
            for cbx in cbx_list:
                check.append({'name': cbx['key'], 'value': 'Yes' if cbx['value'] in input_list else 'No'})
            print('')
            print_table(check, {'Input': 'name', 'Value': 'value'})
            print('')
            if input_yes_no('Are These answers ok?', True):
                confirmed = True

        # Add input to ret_dict
        ret_dict[name] = input_list

    elif type_hint == 'dropdown':
        print(f"\n{desc}")

        if 'source' in var:
            raw_dd_list = context.get(var["source"], [])
            if not isinstance(raw_dd_list, list):
                raw_dd_list = [raw_dd_list]

            dd_list = [{"key": x, "value": x} for x in raw_dd_list]
        else:
            dd_list = var.get('dd_list', [])

        for i in range(0, len(dd_list)):
            print(f"   {i + 1}. {dd_list[i]['key']}")

        valid_response = False
        while not valid_response:
            default_key = ""
            if not default_is_empty:
                default_dd = [x for x in dd_list if x['value'] == default]
                if not len(default_dd):
                    # If there is a default from the context and it's not an option, clear the default
                    default = ""
                else:
                    default_dd = default_dd[0]
                    default_key = f"({default_dd['key']})"
            response = input(f"Please enter line number of selection {default_key}: ")
            if not len(response) and not default_is_empty:
                # Assume default
                response = default_dd["value"]
                valid_response = True
            elif response.isdigit() and '.' not in response:
                response_index = int(response) - 1
                if response_index < 0 or response_index >= len(dd_list):
                    print(f"Please input a number between 1 and {len(dd_list)}")
                else:
                    value = dd_list[response_index]['value']
                    ret_dict[name] = value
                    valid_response = True
            else:
                print("Please input a number")

    elif type_hint == 'list':
        valid_response = False
        if isinstance(default, list):
            if not default_is_empty:
                print("\nDefault list")
                for item in default:
                    print(f"  - {item}")
        while not valid_response:
            response = input(f"{desc} (comma seperated list, blank to accept default): ")
            if len(response):
                rl = [x.strip() for x in response.split(',') if len(x)]
            else:
                rl = default
                valid_response = True
                continue
            print("You entered:")
            if len(rl):
                for item in rl:
                    print(f"  - {item}")
            else:
                print("[]")

            if input_yes_no("Are these values correct?", True):
                ret_dict[name] = rl
                valid_response = True

    elif type_hint == 'ip_address':
        valid_response = False
        while not valid_response:
            response = input(f"{var.get('description', name)} {default_str}: ")
            if not len(response):
                response = default
            if is_ip(response):
                valid_response = True
            else:
                print("Invalid IP address.")
        ret_dict[name] = response

    elif type_hint == 'password':
        valid_response = False
        while not valid_response:
            password = getpass(f"{var.get('description', name)}: ")
            confirm = getpass(f"Confirm - {var.get('description', name)}: ")
            if password == confirm:
                valid_response = True
            else:
                print("Password and confirm do not match!")
        ret_dict[name] = password

    elif type_hint == 'text':

        response = input(f"{var.get('description', name)} {default_str}: ")
        if not len(response) and not default_is_empty:
            ret_dict[name] = default
        else:
            ret_dict[name] = response

    elif type_hint == 'hidden':
        ret_dict[name] = default

    elif type_hint == 'fqdn_or_ip':
        response = input(f"{var.get('description', name)} {default_str}: ")
        if not len(response) and not default_is_empty:
            ret_dict[name] = default
        else:
            ret_dict[name] = response

    else:
        raise ValueError(f'Unsupported input variable type of {type_hint} in var {name}')

    return ret_dict


def expandedHomePath(directory):
    """Returns a local OS formatted directory string"""
    return expanduser("~") + os.path.sep + os.path.sep.join(directory.strip().split('/'))


def merge_children(config, xml):
    """
    Merge a child xml object into an existing xml config object.
    Recursively searches children for any 'entry' style lists and
    merges accordingly, items with no lists are simply replaced.
    """
    print(f"Merging children into {config.tag} from {xml.tag}")

    # All nodes from new XML document
    for xml_child in xml.getchildren():

        # Check if child node has a matching config node
        config_node = config.xpath(xml_child.tag)
        if len(config_node):
            config_node = config_node[0]

            # Node has entry children somewhere
            if xml_child.xpath(".//entry[@name]"):

                # Node has entry immediate children
                if len(xml_child.xpath("./entry[@name]")):
                    for entry_child in xml_child.getchildren():
                        config_node.append(entry_child)
                        print(f"   Appended child {entry_child.tag} {entry_child.get('name')} to {config_node.tag}")

                # Node has entry children, but not immediately
                else:
                    print(f" Recursing over {xml_child.tag}")
                    merge_children(config_node, xml_child)

            # This node has no entry children
            else:

                # Target node has entry children, don't overwrite
                if len(config_node.xpath(".//entry[@name]")):
                    print(f"Ignoring empty node {xml_child.tag} as config has entry children")
                    continue

                # Both source and target have no entry children, replace the node
                config.remove(config_node)
                config_node.append(xml_child)
                print(f"   Replaced node {xml_child.tag} as no entry children were found")

        # Child node does not have a matching config node
        else:
            config.append(xml_child)
            print(f"   Added node {xml_child.tag} due to missing config node")


def merge_into_parent(xpath, config, child_xml):
    """
    Merge a new xml config structure child_xml into config starting
    at xpath. If xpath does not exist, walk back the path until we
    find a common element, and generate the missing structure
    """

    # Find the first level of matching elements
    print(f"Merging new config node {child_xml.tag} into missing parent")
    xpath_elements = xpath.split("/")
    common_xpath = ""
    cursor_element = None
    i = 1
    while i < len(xpath_elements) - 1:
        common_xpath = "/".join(xpath_elements[:-1 * i])
        cursor_element = config.xpath(common_xpath)
        if len(cursor_element) == 1:
            cursor_element = cursor_element[0]
            break
        elif len(cursor_element) > 1:
            raise Exception(f"First level of matching xml in xpath {xpath} produced {len(cursor_element)} nodes at {common_xpath}")
        i += 1
    if not len(common_xpath):
        raise Exception(f"Unable to find a common level of elements for {xpath}")
    print(f"   Found common element at {cursor_element.getroottree().getpath(cursor_element)}")

    # Create the missing gap of XML elements and place new elements inside
    missing_elements = [x for x in xpath.replace(common_xpath, "").split("/") if x]
    for element in missing_elements:
        attr_name = None
        attr_value = None
        if "@" in element and "[" in element:
            element, predicate = element.split("[")
            for c in '[]@ |&"\'':
                predicate = predicate.replace(c, "")
            attr_name, attr_value = predicate.split("=")
        new_element = etree.Element(element.strip())
        if attr_name:
            new_element.set(attr_name, attr_value)
        cursor_element.append(new_element)
        cursor_element = new_element
        print(f"   Added missing element {cursor_element.tag} to cursor")
    for child in child_xml.getchildren():
        cursor_element.append(child)
        print(f"   Added config element {child.tag} to {cursor_element.tag}")


def merge_xml_into_config(xpath, config, child_xml):
    """
    Merge an xml config structure child_xml into config starting at
    the element returned with xpath, which must either specify a unique
    element, or an element that has not yet been created
    """
    found = config.xpath(xpath)

    # If an equivelant xpath was found, merge the children
    if len(found) == 1:
        found = found[0]
        merge_children(found, child_xml)

    # If no node was found, generate missing XML elements from xpath
    elif len(found) == 0:
        merge_into_parent(xpath, config, child_xml)

    else:
        raise Exception("Skillet xpath returned multiple results on device, cannot merge.")


def format_xml_string(xml, indent=0):
    """Take a string of XML and reformat it for proper spacing at a specified indent level"""

    parser = etree.XMLParser(remove_blank_text=True)
    xml_doc = etree.fromstring(xml, parser)
    temp_file = BytesIO()
    xml_doc.getroottree().write(temp_file, pretty_print=True)
    temp_file.seek(0)
    formatted_xml = temp_file.read().decode()
    return "\n".join([" " * indent + x for x in formatted_xml.split("\n")])


def get_password_input(prompt):
    """Get a password input and confirm it, returning the password"""
    match = False
    password = ""
    while not match:
        password = getpass(prompt + ": ")
        confirm = getpass(prompt + " - confirm: ")
        match = password == confirm and len(password) > 7
        if not password == confirm:
            print("Passwords did not match, please try again")
        if len(password) < 8:
            print("Password must be 8 characters")
    return password


def load_config_file(source_name, pan=None):
    """
    Wrapper to load a config file, either off of disk or the device. Panoply session
    is required as a second argument if config file may be on device
    """

    device_configs = ["running", "candidate", "baseline"]  # Can be pulled off of device by name not file
    if source_name.startswith("file:") or pan is None:
        file_name = "".join(source_name.split(":")[1:])
        with open(file_name, "r") as f:
            return f.read()
    elif not pan.connected:
        # offline mode was requested, try to load local files only
        with open(source_name, "r") as f:
            return f.read()
    elif source_name in device_configs or source_name.replace("-", "").isdigit():
        return pan.get_configuration(config_source=source_name)
    else:
        return pan.get_saved_configuration(source_name)


def get_input(var_label: str, var_default: str) -> str:
    """
    utility method to get input from the user and return the default value if nothing is entered from the user

    :param var_label: Label to show to the user
    :param var_default: default to use if nothing is entered
    :return: value entered from the user or default is input is None or ""
    """
    val = input(f"{var_label} <{var_default}>: ")
    if val is None or val == "":
        val = var_default

    return val


def load_app_skillet(skillet_name) -> Skillet:
    """
    Returns a SLI specific application skillet found in the app_skillets folder

    :param skillet_name: Name of skillet to load
    :return: Skillet loaded from name
    """
    sli_path = Path(__file__).parent.joinpath("app_skillets").resolve()
    inline_sl = SkilletLoader(sli_path)
    app_skillet: Skillet = inline_sl.get_skillet_with_name(skillet_name)
    if not app_skillet:
        raise AppSkilletNotFoundException("Could not find required resources")
    return app_skillet



import asyncssh
import re


class AsyncSSHSession:

    BUFFER_LEN = 9999

    def __init__(self, device, username, password):
        self.device = device
        self.username = username
        self.password = password
        self.client = None
        self.c_stdin = None
        self.c_stdout = None
        self.c_stderr = None
        self.prompt = None
        self.hostname = None
        self.has_error = False
        self.error = None

    async def connect(self):
        self.client = await asyncssh.connect(
            self.device,
            username=self.username,
            password=self.password,
            known_hosts=None,
            request_pty="force",
            term_type="vt100"
        )
        self.c_stdin, self.c_stdout, self.c_stderr = await self.client.open_session()
        await self._read_until_hostname()
        await self._set_cli_options()

    async def _read_until_hostname(self):
        """
        Read from stdout until a prompt is found, this will be variable
        length depending on any banners
        """
        prompt_found = None
        while not prompt_found:
            data = await self.c_stdout.read(self.BUFFER_LEN)
            prompt_found = self._check_for_prompt(data)
        self.prompt = data.strip()
        self.hostname = self.prompt.split("@")[1]

    async def _set_cli_options(self):
        """
        Send cli options to make interactive session script friendly
        """
        await self.cli_command("set cli scripting-mode on")
        await self.cli_command("set cli pager off")

    @staticmethod
    def _check_for_prompt(data):
        """
        Check for a line containing a prompt in a block of output
        """
        test_data = data.split("\n")
        if len(test_data):
            return re.match('^.*@.*[#>]$', test_data[-1].strip())
        return None

    def error_check_output(self, output):
        """
        Return True on detecting error in output
        """
        lines = output.split("\n")
        for line in lines:
            if line.startswith("Unknown command:"):
                self.has_error = True
                self.error = line

    async def recv_until_prompt(self, echo=False):
        """
        Receive data from stdin until a prompt is seen, return any text
        """
        prompt_found = None
        data = ''
        while not prompt_found:
            recv_data = await self.c_stdout.read(self.BUFFER_LEN)
            recv_data = recv_data.replace("\r\n", "\n")
            data += recv_data
            prompt_found = self._check_for_prompt(recv_data)

        # Return all lines except the first and last line
        if echo:
            return data
        return "\n".join(data.split("\n")[1:-1])

    async def cli_command(self, command, echo=False):
        """
        Sends a command and returns response as string
        """
        self.has_error = False
        self.c_stdin.write(command + "\n")
        output = await self.recv_until_prompt(echo=echo)
        self.error_check_output(output)
        return output

    async def run_command_script(self, lines, out_file=None):
        """
        Take a list of lines and run them as a script, return True if no error
        Return False if error
        """
        self.has_error = False
        ll = lines
        if isinstance(ll, str):
            ll = ll.split("\n")

        # Setup output options
        session_output = ""
        if out_file:
            session_output = open(out_file, "w")

        for line in ll:
            if not len(line):
                continue
            output = await self.cli_command(line.strip(), echo=True)

            # Save output to file if specified
            if out_file:
                session_output.write(output)
            else:
                session_output += output

            # Stop running the script if an error was found
            self.error_check_output(output)
            if self.has_error:
                if out_file:
                    session_output.close()
                    return False
                return session_output

        # Only return output lines if not saving to a file
        if out_file:
            session_output.close()
            return True
        return session_output



from .base import BaseCommand
import yaml

from skilletlib.skillet.panos import PanosSkillet


class RollupSkillet(BaseCommand):
    sli_command = 'rollup_skillet'
    short_desc = 'Rollup a panos skillet with external XML into a single file'
    no_skillet = True
    no_context = True
    help_text = """

    Usage:
        sli rollup_skillet [skillet-file] [out-file]

"""

    def run(self):
        if not len(self.args) == 2:
            print(self.help_text)
            return

        # Load skillet and copy over non snippet keys
        skillet_file = self.args[0]
        out_file = self.args[1]
        with open(skillet_file, "r") as f:
            skillet = yaml.safe_load(f)

        # Validate supported keys in snippets
        for snippet in skillet["snippets"]:
            for key in snippet.keys():
                if key not in ["name", "xpath", "when", "file"]:
                    print(f"Snippet has unsupported key {key}")
                    return
            for key in ["name", "xpath", "file"]:
                if key not in snippet:
                    print(f"Snippet missing required key {key}")
                    return

        # Load updated dict with merged configuration files
        updated_dict = PanosSkillet(skillet).skillet_dict
        for s in updated_dict["snippets"]:
            del_items = [key for key, value in s.items() if not value or key == "file"]
            for item in del_items:
                del(s[item])

        out_str = PanosSkillet(updated_dict).dump_yaml()
        with open(out_file, "w") as f:
            f.write(out_str)



import json

from .base import BaseCommand


class ShowContext(BaseCommand):
    sli_command = "show_context"
    short_desc = "Print out contents of an existing context"
    no_skillet = True
    no_context = True

    def run(self):
        context_name = self.sli.options.get("context_name", "")
        if len(context_name) < 1:
            context_name = self.args[0] if len(self.args) > 0 else "default"
        context = self.sli.cm.load_context(from_file=context_name)
        if not len(context.keys()):
            print(f"Unable to load context {context_name}")
            return
        if self.sli.no_config and "config" in context:
            context.pop("config")
        hidden_fields = ["TARGET_PASSWORD"]
        for hf in hidden_fields:
            if hf in context:
                context[hf] = "*****"

        if len(self.args) > 1:
            print(json.dumps(context.get(self.args[1]), indent=4))
        else:
            print(json.dumps(context, indent=4))



from sli.decorators import load_variables
from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type
from .base import BaseCommand

"""
Execute a REST skillet and generate results
"""


class ValidateCommand(BaseCommand):
    sli_command = 'rest'
    short_desc = 'Execute a validation skillet of type REST'

    @require_single_skillet
    @require_skillet_type('rest')
    @load_variables
    def run(self):
        """
        Execute a REST skillet
        """

        self.sli.skillet.execute(self.sli.context)



from getpass import getpass

from skilletlib.exceptions import TargetConnectionException
from skilletlib.panoply import Panoply

from .base import BaseCommand
from ..decorators import require_ngfw_connection_params


class ConnectCommand(BaseCommand):
    sli_command = "connect"
    short_desc = "Connect to NGFW and save auth information into context if desired"
    no_skillet = True

    def execute(self):
        self.sli.context.pop("TARGET_IP", "")
        self.sli.context.pop("TARGET_USERNAME", "")
        self.sli.context.pop("TARGET_PASSWORD", "")
        super().execute()

    @require_ngfw_connection_params
    def run(self):

        target_ip = self.sli.context["TARGET_IP"]
        target_username = self.sli.context["TARGET_USERNAME"]
        target_password = self.sli.context["TARGET_PASSWORD"]

        pan = Panoply(target_ip, target_username, target_password)

        wait = self.sli.options.get("wait")
        if wait:
            if not isinstance(wait, int):
                wait = int(wait)

            pan.wait_for_device_ready(30, wait)

        if not pan.connected:
            raise TargetConnectionException("Unable to connect to device...")

        sw_version = pan.facts["sw-version"]
        print(f"Connected to device: {target_ip} running PAN-OS: {sw_version}")




from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type
from .base import BaseCommand

from jinja2 import Environment, meta
import xlsxwriter
import os

"""
Generate output of a panos skillet and write to file
"""


class SpreadsheetCommand(BaseCommand):
    sli_command = 'spreadsheet'
    short_desc = 'Generate a user editable spreadsheet from a template'
    suppress_output = True
    help_text = """

        Load a template skillet, and convert it to a spreadsheet a user
        can edit to manipulate the set commands output

        Usage:
            sli spreadsheet -n template_skillet -o output_directory
"""

    @require_single_skillet
    @require_skillet_type('template')
    def run(self):

        out_directory = self.sli.options.get("out_file", "")

        for snippet in self.sli.skillet.get_snippets():

            # Find out_file
            out_file = snippet.file + ".xlsx"
            file_split = snippet.file.split(".")
            if len(file_split) > 1:
                out_file = out_directory + ".".join(file_split[:-1]) + ".xlsx"

            # Verify input file is a set command template
            with open(self.sli.skillet.path + os.path.sep + snippet.file, "r") as f:
                contents = f.readlines()
            i = 1
            for line in [x for x in contents if len(x.strip()) > 1]:
                if len(line) and not any([line.strip().startswith(x) for x in ["set", "delete", "#"]]):
                    print(f"Problem at line {i}\n  {line.strip()}")
                    print("Invalid file, only configuration mode commands and comments starting with # are allowed")
                    return
                i += 1

            workbook = xlsxwriter.Workbook(out_file)

            # add columns and format width
            worksheet_values = workbook.add_worksheet("values")
            worksheet_values.set_column(0, 0, 30)
            worksheet_values.set_column(1, 1, 30)
            worksheet_set = workbook.add_worksheet("set commands")

            # Add a bold format to use to highlight cells.
            bold = workbook.add_format({"bold": 1})

            # positional list to map variables into formula
            # padded since variables start at row 2 with zero offset
            variable_list = ['first row', 'second row']

            worksheet_values.write(0, 0, 'Variable Name', bold)
            worksheet_values.write(0, 1, 'Variable Value', bold)

            # Add variables to workbook
            row = 1
            for variable in self.sli.skillet.variables:
                worksheet_values.write(row, 0, variable['name'])
                worksheet_values.write(row, 1, variable['default'])
                worksheet_values.write(row, 2, variable['description'])
                variable_list.append(variable['name'])
                row += 1
            row = 1

            # Add set commands to workbook
            for line in contents:

                # Read the line and create a variable set
                env = Environment()
                var_set = sorted(meta.find_undeclared_variables(env.parse(line)))

                # Check if line has a single variable
                if len(var_set) == 1:
                    # Replace quote with 2xquotes to be excel friendly
                    line = line.replace('"', '""').strip()
                    var_name = var_set[0]
                    jinja_var = "{{ " + var_name + " }}"
                    cell_pos = variable_list.index(var_name)

                    # Formula format  = =SUBSTITUTE("set deviceconfig system dns-setting servers primary {0}", "{0}", '0-Config Values'!B21)
                    form_line = f'=SUBSTITUTE("{line}", "{jinja_var}", \'values\'!B{cell_pos})'
                    worksheet_set.write(row, 0, form_line)

                # Check if line has 2 variables when requires nested SUBSTITUTEs
                elif len(var_set) == 2:
                    # Replace quote with 2xquotes to be excel friendly
                    line = line.replace('"', '""').strip()
                    var_name_0 = var_set[0]
                    jinja_var_0 = "{{ " + var_name_0 + " }}"
                    cell_pos_0 = variable_list.index(var_name_0)

                    var_name_1 = var_set[1]
                    jinja_var_1 = "{{ " + var_name_1 + " }}"
                    cell_pos_1 = variable_list.index(var_name_1)

                    # Inner is first substitute and all is nested substitute for excel
                    form_line_inner = f'SUBSTITUTE("{line}", "{jinja_var_0}", \'values\'!B{cell_pos_0})'
                    form_line_all = f'=SUBSTITUTE({form_line_inner}, "{jinja_var_1}", \'values\'!B{cell_pos_1})'
                    worksheet_set.write(row, 0, form_line_all)

                else:
                    # Replace quotes and also remove hidden quotes for standard non-formula cells
                    line = line.replace('"', '""').strip()
                    worksheet_set.write(row, 0, line)

                row += 1

            workbook.close()
            print(f"Workbook completed and saved to {out_file}")



from .base import BaseCommand
from sli.decorators import require_ngfw_connection_params, require_panoply_connection
import requests
from lxml import etree
import os


class LoadConfig(BaseCommand):
    sli_command = "load_config"
    short_desc = "Load an XML configuration file onto an NGFW as candidate"
    no_skillet = True
    help_text = """
        Usage:
            sli load_config config_file.xml
    """

    @require_ngfw_connection_params
    @require_panoply_connection
    def run(self, pan):

        if not len(self.args) == 1:
            print(self.help_text)
            return
        file_name = self.args[0]
        if not file_name.endswith(".xml"):
            print("Target file must be an xml file")
            return

        # Push configuration file to the device
        print(f"Loading configuration file {file_name}")
        with open(file_name, 'rb') as f:
            params = {
                "type": "import",
                "category": "configuration",
                "key": pan.xapi.api_key
            }
            r = requests.post(
                f"https://{self.sli.context['TARGET_IP']}/api",
                params=params,
                verify=False,
                files={"file": f}
            )
        if not r.status_code == 200:
            print(f"Unable to load configuration on firewall, {r.status_code} received")
        print(f"Loaded {file_name}")
        xml = etree.fromstring(r.text)
        if not xml.xpath("@status")[0].lower() == "success":
            err = "".join(xml.xpath("//msg/text()"))
            if not len(err):
                err = "".join(xml.xpath("//line/text()"))
            if len(err):
                print(f"Error from device: \n{err}")
            else:
                print("Unable to push configuration to device")
            return

        # Configuration file is on device, load it as candidate config
        file_name = file_name.replace('/', os.path.sep).split(os.path.sep)[-1]
        xml = f"<load><config><from>{file_name}</from></config></load>"
        r = pan.execute_op(xml)
        xml = etree.fromstring(r)
        msg = xml.xpath("//line/text()")[0]
        if len(msg):
            print(msg)
        else:
            print("Unable to retrieve op results from device")



import os
from pathlib import Path

import yaml
from git import Repo
from skilletlib import Skillet
from skilletlib import SkilletLoader

from .diff import DiffCommand
from ..errors import InvalidArgumentsException
from ..errors import SLIException
from ..tools import get_input


class AnsibleRoleCommand(DiffCommand):
    sli_command = "ansible_role"
    short_desc = (
        "Builds an Ansible Role from the differences between two config versions, candidate, running, "
        "previous running, etc"
    )
    no_skillet = True
    repo_url = "https://gitlab.com/panw-gse/as/pan_community-ansible-collection-skeleton.git"
    help_text = """
        ansible_role module requires 0 to 2 arguments.

        - 0 arguments: Diff running config from previous running config
        - 1 argument: Diff previous config or named config against specified running config
        - 2 arguments: Diff first arg named config against second arg named config

        A named config can be either a stored config, candidate, running or a number.
        Positive numbers must be used to specify iterations, 1 means 1 config revision ago

        Example: Get diff between running config and previous running config

            user$ sli ansible_role 1 -uc  -od /tmp/roles

        Example: Get diff between running config and the candidate config

            user$ sli ansible_role running candidate -uc -od /tmp/roles

        Example: Get the diff of all changes made to this device using the autogenerated 'baseline' config, which is
        essentially blank.

            user$ sli ansible_role baseline -uc -od /tmp/roles

        Example: Get the diff between the second and third most recent running configs

            user$ sli ansible_role 3 2 -uc -od /tmp/roles
    """

    @staticmethod
    def __write_file_to_location(file_path_root: Path, file_name: str, file_contents: str) -> None:
        """
        utility method to ensure a path exists and write the file_contents into file_name at file_path_root

        :param file_path_root: directory in which to check for the existence of file_name
        :param file_name: name of the file to write
        :param file_contents: string contents of the file to write
        :return: None
        """

        file_path = file_path_root.joinpath(file_name).resolve()

        if not file_path_root.exists():
            raise SLIException("Could not write file!")

        try:
            with file_path.open(mode="w") as fp:
                fp.write(file_contents)
        except OSError as oe:
            raise SLIException(f"Could not write file! {oe}")

    @staticmethod
    def _load_skillet(skillet_name: str, source_dir: Path) -> Skillet:
        """
        Returns a skillet found in the source_dir folder
        """
        inline_sl = SkilletLoader(source_dir)
        app_skillet: Skillet = inline_sl.get_skillet_with_name(skillet_name)

        if not app_skillet:
            raise SLIException("Could not find required resources")

        return app_skillet

    def _get_vars(self) -> list:
        """
        Return a list of a single variable to populate the test, readme, and variable_list files
        """
        return [
            {
                "name": "custom_variable",
                "description": "Custom Variable",
                "default": "default_value",
                "type_hint": "text",
            }
        ]

    def _get_version(self) -> str:
        return self.pan.facts["sw-version"]

    def _handle_snippets(self, snippets: list, variable_list: list) -> None:
        """
         override method to perform action on the list of diffs. In this case, build the ansible role structure

        :param snippets: list of differences found from Skilletlib
        "param variable_list: list of variables as found in the snippets
        :return: None
        """
        if not len(snippets):
            print("No Configuration diffs found! Try a different combination of configuration sources")
            exit(1)

        namespace = get_input("Namespace", "pan_community")
        role_name = get_input("Role Name", "generated_role")
        description = get_input("Role Description", "generated_role")
        author_name = get_input("Author Name", "john doe")

        # get version and major version from connected NGFW
        version = self._get_version()
        major_version = ".".join(version.split(".")[0:2])

        # verify output dir structure and existence
        output_dir: str = self.sli.options.get("output_directory", None)

        if output_dir is None:
            output_dir = "./ansible_collections"

        if str(output_dir).startswith("./"):
            output_path = Path(os.getcwd()).joinpath(output_dir).resolve()
        else:
            output_path = Path(output_dir)
            if not output_path.exists() and not output_path.parent.exists():
                raise InvalidArgumentsException("Output directory does not exist!")

        if output_dir != "ansible_collections":
            output_path = output_path.joinpath("ansible_collections")
            output_path.mkdir(parents=True, exist_ok=True)

        if next(os.scandir(output_path), None):
            raise SLIException(f"Ansible Roles Directory: {output_path} must be empty!")

        # clone our skeleton repo
        Repo.clone_from(url=self.repo_url, to_path=output_path, env={"GIT_SSL_NO_VERIFY": "1"})

        # rename the directory structure appropriately
        # it comes from the skeleton repo as
        # example/
        #   skeleton/
        #       roles/
        #           skeleton/

        # rename example to namespace
        top_level = output_path.joinpath("example")
        namespace_path = output_path.joinpath(namespace)
        top_level.rename(namespace_path)

        # rename skeleton to role_name
        skeleton = namespace_path.joinpath("skeleton")
        collection_path = skeleton.parent.joinpath(role_name)
        skeleton.rename(collection_path)

        # rename roles/skeleton to roles/role_name
        skeleton_role_path = collection_path.joinpath("roles/skeleton")
        role_path = skeleton_role_path.parent.joinpath(role_name)
        skeleton_role_path.rename(role_path)

        # create our inline skilletloader instance
        inline_sl = SkilletLoader(output_path)

        # create README
        readme_skillet = inline_sl.get_skillet_with_name("readme_template")
        readme_outout = readme_skillet.execute(
            {
                "role_name": role_name,
                "description": description,
                "author_name": author_name,
                "namespace": namespace,
                "variable_list": variable_list,
            }
        )

        self.__write_file_to_location(output_path, "README.md", readme_outout["template"])
        self.__write_file_to_location(collection_path, "README.md", readme_outout["template"])

        # create Role README
        role_readme_skillet = inline_sl.get_skillet_with_name("role_readme_template")
        role_readme_outout = role_readme_skillet.execute(
            {
                "role_name": role_name,
                "description": description,
                "author_name": author_name,
                "namespace": namespace,
                "variable_list": variable_list,
            }
        )

        self.__write_file_to_location(role_path, "README.md", role_readme_outout["template"])

        # write out galaxy.yml file from template
        galaxy_skillet = inline_sl.get_skillet_with_name("ansible_galaxy_template")

        output = galaxy_skillet.execute(
            {"role_name": role_name, "description": description, "author_name": author_name}
        )

        self.__write_file_to_location(collection_path, "galaxy.yml", output["template"])

        # add major version to supported_versions list
        supported_versions = [major_version]

        # create role vars file
        vars_file_data = dict()
        vars_file_data["supported_versions"] = supported_versions
        supported_versions_yaml = yaml.safe_dump(vars_file_data)

        role_vars_path = role_path.joinpath("vars")
        self.__write_file_to_location(role_vars_path, "main.yml", supported_versions_yaml)

        # create version specific task file
        role_tasks_path = role_path.joinpath("tasks")
        ansible_task_template = inline_sl.get_skillet_with_name("ansible_task_template")
        task_output = ansible_task_template.execute({"snippets": snippets})
        self.__write_file_to_location(role_tasks_path, f"panos-{major_version}.yml", task_output["template"])

        # create role meta file
        role_meta_path = role_path.joinpath("meta")
        ansible_task_meta_template = inline_sl.get_skillet_with_name("ansible_task_meta_template")
        task_output = ansible_task_meta_template.execute(
            {
                "role_name": role_name,
                "description": description,
                "author_name": author_name,
                "variable_list": variable_list,
            }
        )
        self.__write_file_to_location(role_meta_path, "main.yml", task_output["template"])

        # create role test file
        role_test_path = role_path.joinpath("tests")
        ansible_test_template = inline_sl.get_skillet_with_name("ansible_test_template")
        task_output = ansible_test_template.execute({"role_name": role_name, "variable_list": variable_list})
        self.__write_file_to_location(role_test_path, "test.yml", task_output["template"])

        print(f"Ansible Role created in {output_path} successfully!")



from getpass import getpass
from skilletlib.panoply import Panoply
from sli.tools import get_var
from sli.ssh import SSHSession
from lxml import etree
from skilletlib.exceptions import LoginException, TargetConnectionException


def require_ngfw_connection_params(func):
    """
    Validate that SLI command has appropriate contexet paramaters
    to facilitate a connection to an NGFW
    """

    def wrap(command):
        ensure_ngfw_connection_params(command)
        return func(command)

    return wrap


def require_panoply_connection(func):
    """
    Requires a connected panoply session to the target device, will pass in
    the connected panoply instance as an argument
    """

    def wrap(command):
        try:
            pan = get_panoply_from_context(command)
            return func(command, pan)
        except (LoginException, TargetConnectionException) as e:
            print(f"Login error: {e}")

    return wrap


def require_config(func):
    """
    Decorator to require a command to have an etree representing a panos
    configuration. User can either specify a file to load, or one will be
    retrieved from a device by default
    """

    def wrap(command):
        config = None
        config_file = command.sli.options.get("config_file")
        parser = etree.XMLParser(remove_blank_text=True)
        # Load configuration from file
        if config_file:
            with open(config_file, "r") as f:
                contents = f.read()
                command.sli.context["config"] = contents
                config = etree.fromstring(contents, parser).getroottree()
        # No configuration file specified, connect to device
        else:
            ensure_ngfw_connection_params(command)
            pan = get_panoply_from_context(command)
            contents = pan.get_configuration()
            command.sli.context["config"] = contents
            config = etree.fromstring(contents, parser).getroottree()
        return func(command, config)

    return wrap


def require_ngfw_ssh_session(func):
    """
    Decorator to require a connected SSH paramiko session be passed to
    the calling command. The SSH session will already have an invoked shell
    """

    def wrap(command):
        print(f"Connecting to {command.sli.context['TARGET_IP']}...")
        ssh = SSHSession(
            command.sli.context["TARGET_IP"],
            username=command.sli.context["TARGET_USERNAME"],
            password=command.sli.context["TARGET_PASSWORD"],
        )
        print("Connected.")
        return func(command, ssh)

    return wrap


def load_variables(func):
    """Load variables from skillet and get user input if not supplied"""

    def wrap(command):
        if command.sli.options.get("defaults") is True:
            print("Specified accept defaults. Skipping user prompts with default options...")
        for var in command.sli.skillet.variables:
            get_var(var, command.raw_args, command.sli.context, options=command.sli.options)

        return func(command)

    return wrap


def require_single_skillet(func):
    """Commands decorated with this require one skillet to be uniquely specified"""

    def wrap(command):
        if not command.sli.options.get("name") and len(command.sli.skillets) > 1:
            print(
                f"Specify a skillet to run with --name or -n when more than 1 is present for command {command.sli_command}"
            )
            exit(1)

        target_name = (
            command.sli.options.get("name") if command.sli.options.get("name") else command.sli.skillets[0].name
        )
        command.sli.skillet = command.sli.sl.get_skillet_with_name(target_name)
        if command.sli.skillet is None:
            print(f"Unable to load skillet {target_name} by name")
            exit(1)
        return func(command)

    return wrap


def require_skillet_type(*args):
    """
    Commands decorated with this require a specific type of skillet to be executed.
    Inputs expected as *args, selected skillet type must match one
    """

    def inner(func):
        def wrap(command):
            skillet_type = command.sli.skillet.type
            if skillet_type not in args:
                print(
                    f'Invalid type of skillet ({skillet_type}) for command {command.sli_command}, requires one of: {", ".join(args)}'
                )
                exit(1)
            return func(command)

        return wrap

    return inner


def ensure_ngfw_connection_params(command):
    """
    Support function for decorators designed to ensure that
    device connection parameters are present in the context
    """
    offline_mode = command.sli.options.get("offline", False)
    if offline_mode:
        return

    # map CLI options into context dict
    device = command.sli.options.get("device", "")
    username = command.sli.options.get("username", "")
    password = command.sli.options.get("password", "")
    if len(device) > 0:
        command.sli.context["TARGET_IP"] = device
    if len(username) > 0:
        command.sli.context["TARGET_USERNAME"] = username
    if len(password) > 0:
        command.sli.context["TARGET_PASSWORD"] = password

    # If we don't have required input parameters, get them from the user
    if len(command.sli.context.get("TARGET_IP", "")) < 1:
        command.sli.context["TARGET_IP"] = input("Device: ")
    if len(command.sli.context.get("TARGET_USERNAME", "")) < 1:
        command.sli.context["TARGET_USERNAME"] = input("Username: ")
    if len(command.sli.context.get("TARGET_PASSWORD", "")) < 1:
        command.sli.context["TARGET_PASSWORD"] = getpass()


def get_panoply_from_context(command):
    """
    Support function for decorators to retrieve a connected panoply
    session instantiated from the context
    """

    # check for offline mode
    offline = command.sli.options.get("offline", False)
    if offline:
        pan = Panoply()
        return pan

    context = command.sli.context
    api_port = int(context["TARGET_PORT"]) if "TARGET_PORT" in context else 443
    pan = Panoply(context["TARGET_IP"], context["TARGET_USERNAME"], context["TARGET_PASSWORD"], api_port=api_port)

    wait = command.sli.options.get("wait")
    if wait:
        if not isinstance(wait, int):
            wait = int(wait)

        pan.wait_for_device_ready(30, wait)

    if not pan.connected:
        raise TargetConnectionException("Unable to connect to device")
    return pan



from .base import BaseCommand
from sli.decorators import require_ngfw_connection_params, require_panoply_connection
from sli.tools import load_config_file


class ConfigureCommand(BaseCommand):
    sli_command = "save_config"
    short_desc = "Execute a configuration skillet of type panos"
    no_skillet = True
    help_text = """
    Save a configuration file off a device locally. config_on_device may be
    running, candidate, or any config file saved on a device. Specify -o config.xml
    to save off as config.xml, or the configuration will just be printed to the screen

    Usage:

        sli save_config config_on_device -o config.xml
"""

    @require_ngfw_connection_params
    @require_panoply_connection
    def run(self, pan):

        if len(self.args) != 1:
            print(self.help_text)
            return

        source_name = self.args[0]
        config = load_config_file(source_name, pan=pan)
        out_file = self.sli.options.get("out_file")
        if out_file:
            with open(out_file, "w") as f:
                f.write(config)
            print(f"Configuration file saved to {out_file}")
        else:
            print(config)



from .base import BaseCommand
from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type

from skilletlib.skillet.panos import PanosSkillet


class RollupPlaylist(BaseCommand):
    sli_command = 'rollup_playlist'
    short_desc = 'Rollup a panos skillet in playlist format into a single file'
    no_context = True
    help_text = """

    Usage:
        sli rollup_playlist -n [skillet] [out-file]

"""

    @require_single_skillet
    @require_skillet_type("panos", "panorama")
    def run(self):
        if not len(self.args) == 1:
            print(self.help_text)
            return
        out_file = self.args[0]

        new_skillet = {}
        for attr in ["name", "label", "description", "type", "labels", "variables"]:
            new_skillet[attr] = getattr(self.sli.skillet, attr)
        new_skillet["name"] += "_rollup"
        new_skillet["snippets"] = [x.metadata for x in self.sli.skillet.snippets]
        for snippet in new_skillet["snippets"]:
            del_keys = [key for key, value in snippet.items() if not value]
            for key in del_keys:
                snippet.pop(key)

        out_str = PanosSkillet(new_skillet).dump_yaml()
        with open(out_file, "w") as f:
            f.write(out_str)
        print(f"New skillet written to {out_file}")
